""" HTTP API methods for Dagobah daemon. """

import StringIO
import json

from flask import request, abort, send_file
from flask_login import login_required

from .daemon import app
from .util import validate_dict, api_call, allowed_file

dagobah = app.config['dagobah']

@app.route('/api/jobs', methods=['GET'])
@login_required
@api_call
def get_jobs():
    return dagobah._serialize().get('jobs', {})


@app.route('/api/job', methods=['GET'])
@login_required
@api_call
def get_job():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    if not job:
        abort(400)

    return job._serialize()


@app.route('/api/logs', methods=['GET'])
@login_required
@api_call
def get_run_log_history():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)
    return task.get_run_log_history()


@app.route('/api/log', methods=['GET'])
@login_required
@api_call
def get_log():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name', 'task_name', 'log_id'],
                         job_name=str,
                         task_name=str,
                         log_id=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)
    return task.get_run_log(args['log_id'])


@app.route('/api/head', methods=['GET'])
@login_required
@api_call
def head_task():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str,
                         stream=str,
                         num_lines=int):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)

    call_args = {}
    for key in ['stream', 'num_lines']:
        if key in args:
            call_args[key] = args[key]
    return task.head(**call_args)


@app.route('/api/tail', methods=['GET'])
@login_required
@api_call
def tail_task():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str,
                         stream=str,
                         num_lines=int):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)

    call_args = {}
    for key in ['stream', 'num_lines']:
        if key in args:
            call_args[key] = args[key]
    return task.tail(**call_args)


@app.route('/api/add_job', methods=['POST'])
@login_required
@api_call
def add_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    dagobah.add_job(args['job_name'])


@app.route('/api/delete_job', methods=['POST'])
@login_required
@api_call
def delete_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    dagobah.delete_job(args['job_name'])


@app.route('/api/start_job', methods=['POST'])
@login_required
@api_call
def start_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.start()


@app.route('/api/retry_job', methods=['POST'])
@login_required
@api_call
def retry_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.retry()


@app.route('/api/add_task_to_job', methods=['POST'])
@login_required
@api_call
def add_task_to_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_command', 'task_name'],
                         job_name=str,
                         task_command=str,
                         task_name=str,
                         task_target=str):
        abort(400)

    dagobah.add_task_to_job(args['job_name'],
                            args['task_command'],
                            args['task_name'],
                            hostname=args.get("task_target", None))


@app.route('/api/delete_task', methods=['POST'])
@login_required
@api_call
def delete_task():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.delete_task(args['task_name'])


@app.route('/api/add_dependency', methods=['POST'])
@login_required
@api_call
def add_dependency():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name',
                                   'from_task_name',
                                   'to_task_name'],
                         job_name=str,
                         from_task_name=str,
                         to_task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.add_dependency(args['from_task_name'], args['to_task_name'])

@app.route('/api/delete_dependency', methods=['POST'])
@login_required
@api_call
def delete_dependency():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name',
                                   'from_task_name',
                                   'to_task_name'],
                         job_name=str,
                         from_task_name=str,
                         to_task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.delete_dependency(args['from_task_name'], args['to_task_name'])


@app.route('/api/schedule_job', methods=['POST'])
@login_required
@api_call
def schedule_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'cron_schedule'],
                         job_name=str,
                         cron_schedule=str):
        abort(400)

    if args['cron_schedule'] == '':
        args['cron_schedule'] = None

    job = dagobah.get_job(args['job_name'])
    job.schedule(args['cron_schedule'])


@app.route('/api/stop_scheduler', methods=['POST'])
@login_required
@api_call
def stop_scheduler():
    dagobah.scheduler.stop()


@app.route('/api/restart_scheduler', methods=['POST'])
@login_required
@api_call
def restart_scheduler():
    dagobah.scheduler.restart()


@app.route('/api/terminate_all_tasks', methods=['POST'])
@login_required
@api_call
def terminate_all_tasks():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.terminate_all()


@app.route('/api/kill_all_tasks', methods=['POST'])
@login_required
@api_call
def kill_all_tasks():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.kill_all()


@app.route('/api/terminate_task', methods=['POST'])
@login_required
@api_call
def terminate_task():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)
    task.terminate()


@app.route('/api/kill_task', methods=['POST'])
@login_required
@api_call
def kill_task():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)
    task.kill()


@app.route('/api/edit_job', methods=['POST'])
@login_required
@api_call
def edit_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str,
                         name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    del args['job_name']
    job.edit(**args)


@app.route('/api/update_job_notes', methods=['POST'])
@login_required
@api_call
def update_job_notes():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'notes'],
                         job_name=str,
                         notes=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.update_job_notes(args['notes'])


@app.route('/api/edit_task', methods=['POST'])
@login_required
@api_call
def edit_task():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str,
                         name=str,
                         command=str,
                         soft_timeout=int,
                         hard_timeout=int,
                         hostname=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)

    # validate host
    if 'hostname' in args and args.get('hostname') not in dagobah.get_hosts():
        # Check for empty host, if so then task is no longer remote
        if not args.get('hostname'):
            args['hostname'] = None
        else:
            abort(400)

    del args['job_name']
    del args['task_name']
    job.edit_task(task.name, **args)


@app.route('/api/set_soft_timeout', methods=['POST'])
@login_required
@api_call
def set_soft_timeout():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name', 'soft_timeout'],
                         job_name=str,
                         task_name=str,
                         soft_timeout=int):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)

    task.set_soft_timeout(args['soft_timeout'])


@app.route('/api/set_hard_timeout', methods=['POST'])
@login_required
@api_call
def set_hard_timeout():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name', 'hard_timeout'],
                         job_name=str,
                         task_name=str,
                         hard_timeout=int):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)

    task.set_hard_timeout(args['hard_timeout'])


@app.route('/api/export_job', methods=['GET'])
@login_required
def export_job():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])

    to_send = StringIO.StringIO()
    to_send.write(json.dumps(job._serialize(strict_json=True)))
    to_send.write('\n')
    to_send.seek(0)

    return send_file(to_send,
                     attachment_filename='%s.json' % job.name,
                     as_attachment=True)


@app.route('/api/import_job', methods=['POST'])
@login_required
@api_call
def import_job():
    file = request.files['file']
    if (file and allowed_file(file.filename, ['json'])):
        dagobah.add_job_from_json(file.read(), destructive=True)


@app.route('/api/hosts', methods=['GET'])
@login_required
@api_call
def get_hosts():
    return dagobah.get_hosts()



""" Views for Dagobah daemon. """

from flask import render_template, redirect, url_for, abort
from flask_login import login_required

from .daemon import app
from .api import get_jobs, import_job

dagobah = app.config['dagobah']


@app.route('/', methods=['GET'])
def index_route():
    """ Redirect to the dashboard. """
    return redirect(url_for('jobs'))

@app.route('/jobs', methods=['GET'])
@login_required
def jobs():
    """ Show information on all known Jobs. """
    return render_template('jobs.html',
                           jobs=get_jobs())

@app.route('/jobs/import', methods=['POST'])
@login_required
def jobs_import_view():
    """ Import a Job and redirect to the Jobs page. """
    import_job()
    return redirect(url_for('jobs'))

@app.route('/job/<job_id>', methods=['GET'])
@login_required
def job_detail(job_id=None):
    """ Show a detailed description of a Job's status. """
    jobs = [job for job in get_jobs() if str(job['job_id']) == job_id]
    if not jobs:
        abort(404)
    return render_template('job_detail.html', job=jobs[0], hosts=dagobah.get_hosts())

@app.route('/job/<job_id>/<task_name>', methods=['GET'])
@login_required
def task_detail(job_id=None, task_name=None):
    """ Show a detailed description of a specific task. """
    jobs = get_jobs()
    job = [job for job in jobs if str(job['job_id']) == job_id][0]
    return render_template('task_detail.html',
                           job=job,
                           task_name=task_name,
                           task=[task for task in job['tasks']
                                 if task['name'] == task_name][0])

@app.route('/job/<job_id>/<task_name>/<log_id>', methods=['GET'])
@login_required
def log_detail(job_id=None, task_name=None, log_id=None):
        """ Show a detailed description of a specific log. """
        jobs = get_jobs()
        job = [job for job in jobs if str(job['job_id']) == job_id][0]
        return render_template('log_detail.html',
                               job=job,
                               task_name=task_name,
                               task=[task for task in job['tasks']
                                     if task['name'] == task_name][0],
                               log_id=log_id)

@app.route('/settings', methods=['GET'])
@login_required
def settings_view():
    """ View for managing app-wide configuration. """
    return render_template('settings.html')



import logging

from .text import TextEmail
from .basic import BasicEmail

def get_email_handler(handler_name, email_options):

    if isinstance(handler_name, str):
        handler_name = handler_name.lower()

    #Assume authentication is required if missing for backward compatibility.
    auth_required = email_options.get('auth_required', True)

    user = email_options.get('user', None)
    user = user.lower() if isinstance(user, str) else user

    if handler_name is None:
        return None

    elif auth_required and user is None:
        logging.warn('Email.auth_required is True but Email.user is None. Emailing of reports will be disabled.')
        return None

    elif handler_name == 'text':
        return TextEmail(**email_options)

    elif handler_name == 'basic':
        return BasicEmail(**email_options)

    logging.warn("Dagobahd.email config key unrecognized. Emailing of reports will be disabled.")
    return None



""" Text email template for sending out Dagobah communications. """

from datetime import datetime
from email.mime.text import MIMEText

from .common import EmailTemplate

class TextEmail(EmailTemplate):

    def send_job_completed(self, data):
        self.message = MIMEText(self._job_to_text(data))
        self._construct_and_send('Job Completed: %s' % data.get('name', None))


    def send_job_failed(self, data):
        self.message = MIMEText(self._job_to_text(data))
        self._construct_and_send('Job Failed: %s' % data.get('name', None))


    def send_task_failed(self, data):
        self.message = MIMEText(self._task_to_text(data))
        self._construct_and_send('Task Failed: %s' % data.get('name', None))


    def _task_to_text(self, task):
        """ Return a standard formatting of a Task serialization. """

        started = self._format_date(task.get('started_at', None))
        completed = self._format_date(task.get('completed_at', None))

        success = task.get('success', None)
        success_lu = {None: 'Not executed', True: 'Success',
                      False: 'Failed'}

        run_log = task.get('run_log', {})
        return '\n'.join(['Task: %s' % task.get('name', None),
                          'Command: %s' % task.get('command', None),
                          'Result: %s' % success_lu[success],
                          'Started at: %s' % started,
                          'Completed at: %s' % completed,
                          'Return Code: %s' % run_log.get('return_code', None),
                          'Stdout: %s' % run_log.get('stdout', None),
                          'Stderr: %s' % run_log.get('stderr', None)])


    def _job_to_text(self, job):
        """ Return a standard formatting of a Job serialization. """

        next_run = self._format_date(job.get('next_run', None))

        tasks = ''
        for task in job.get('tasks', []):
            tasks += self._task_to_text(task)
            tasks += '\n\n'

        return '\n'.join(['Job name: %s' % job.get('name', None),
                          'Cron schedule: %s' % job.get('cron_schedule', None),
                          'Next run: %s' % next_run,
                          '',
                          'Parent ID: %s' % job.get('parent_id', None),
                          'Job ID: %s' % job.get('job_id', None),
                          '',
                          'Tasks Detail',
                          '',
                          tasks])


    def _format_date(self, in_date):
        if (not in_date) or (not isinstance(in_date, datetime)):
            return in_date
        return in_date.strftime('%Y-%m-%d %H:%M:%S UTC')



""" Component classes used by core classes. """

import inspect
import logging
from datetime import datetime
from collections import defaultdict
import time
import threading
import json


class EventHandler(object):
    """ Provides an event model for Dagobah methods.

    The Dagobah instance emits events, which can then trigger
    handlers that are registered in this class.
    """

    def __init__(self):
        self.handlers = defaultdict(list)


    def emit(self, event, event_params={}):
        try:
            methods = self.handlers.get(event, [])
            for method, args, kwargs in methods:
                argspec = inspect.getargspec(method)
                if ('event_params' in argspec.args or argspec.keywords is not None):
                    kwargs = dict(kwargs.items() +
                                  {'event_params': event_params}.items())
                method.__call__(*args, **kwargs)
        except Exception:
            logging.exception('Exception emitting event %s' % event)


    def register(self, event, method, *args, **kwargs):
        if 'event_params' in kwargs:
            raise ValueError('event_params is a reserved key')
        self.handlers[event].append((method, args, kwargs))


    def deregister(self, event, method):
        for idx, registered in enumerate(self.handlers[event]):
            if registered[0] == method:
                self.handlers[event].pop(idx)
                break


class JobState(object):
    """ Stores state and related state metadata of a current job. """

    def __init__(self):
        self.status = None

        self.perms = {'allow_start': ['waiting', 'failed'],
                      'allow_change_graph': ['waiting', 'failed'],
                      'allow_change_schedule': ['waiting', 'running', 'failed'],
                      'allow_edit_job': ['waiting', 'failed'],
                      'allow_edit_task': ['waiting', 'failed']}

        for key in self.perms.keys():
            setattr(self, key, None)


    def set_status(self, status):
        status = status.lower()
        if status not in ['waiting', 'running', 'failed']:
            raise ValueError('unknown status %s' % status)

        self.status = status
        self._set_permissions()


    def _set_permissions(self):
        for perm, states in self.perms.iteritems():
            setattr(self, perm, True if self.status in states else False)


class Scheduler(threading.Thread):
    """ Monitoring thread to kick off Jobs at their scheduled times. """

    def __init__(self, parent_dagobah):
        super(Scheduler, self).__init__()
        self.parent = parent_dagobah
        self.stopped = False

        self.last_check = datetime.utcnow()


    def __repr__(self):
        return '<Scheduler for %s>' % self.parent


    def stop(self):
        """ Stop the monitoring loop without killing the thread. """
        self.stopped = True


    def restart(self):
        """ Restart the monitoring loop. """
        self.last_check = datetime.utcnow()
        self.stopped = False


    def run(self):
        """ Continually monitors Jobs of the parent Dagobah. """
        while not self.stopped:
            now = datetime.utcnow()
            for job in self.parent.jobs:
                if not job.next_run:
                    continue
                if job.next_run >= self.last_check and job.next_run <= now:
                    if job.state.allow_start:
                        job.start()
                    else:
                        job.next_run = job.cron_iter.get_next(datetime)
            self.last_checked = now
            time.sleep(1)


class StrictJSONEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            from bson import ObjectId
            if isinstance(o, ObjectId):
                return str(o)
        except ImportError:
            pass
        if isinstance(o, datetime):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)



import os
from pkg_resources import resource_string

from .core import *

def print_standard_conf():
    """ Print the sample config file to stdout. """
    print return_standard_conf()

def return_standard_conf():
    """ Return the sample config file. """
    result = resource_string(__name__, 'daemon/dagobahd.yml')
    result = result % {'app_secret': os.urandom(24).encode('hex')}
    return result



""" HTTP Daemon implementation for Dagobah service. """

import os
import sys
import logging

from flask import Flask, send_from_directory
from flask_login import LoginManager
import yaml

from .. import return_standard_conf
from ..core import Dagobah, EventHandler
from ..email import get_email_handler

app = Flask(__name__)

login_manager = LoginManager()
login_manager.login_view = "login"

location = os.path.realpath(os.path.join(os.getcwd(),
                                         os.path.dirname(__file__)))

class NullHandler(logging.Handler):
    def emit(self, record):
        pass

def replace_nones(dict_or_list):
    """Update a dict or list in place to replace
    'none' string values with Python None."""

    def replace_none_in_value(value):
        if isinstance(value, basestring) and value.lower() == "none":
            return None
        return value

    items = dict_or_list.iteritems() if isinstance(dict_or_list, dict) else enumerate(dict_or_list)

    for accessor, value in items:
        if isinstance(value, (dict, list)):
            replace_nones(value)
        else:
            dict_or_list[accessor] = replace_none_in_value(value)

def get_config_file():
    """ Return the loaded config file if one exists. """

    # config will be created here if we can't find one
    new_config_path = os.path.expanduser('~/.dagobahd.yml')

    config_dirs = ['/etc',
                   os.path.expanduser('~')]
    config_filenames = ['dagobahd.yml',
                        'dagobahd.yaml',
                        '.dagobahd.yml',
                        '.dagobahd.yaml']

    for directory in config_dirs:
        for filename in config_filenames:
            try:
                if os.path.isfile(os.path.join(directory, filename)):
                    to_load = open(os.path.join(directory, filename))
                    config = yaml.load(to_load.read())
                    to_load.close()
                    replace_nones(config)
                    return config
            except:
                pass

    # if we made it to here, need to create a config file
    # double up on notifications here to make sure first-time user sees it
    print 'Creating new config file in home directory'
    logging.info('Creating new config file in home directory')
    new_config = open(new_config_path, 'w')
    new_config.write(return_standard_conf())
    new_config.close()

    new_config = open(new_config_path, 'r')
    config = yaml.load(new_config.read())
    new_config.close()
    replace_nones(config)
    return config


def configure_requests_logger(config, app):
    logger = logging.getLogger('werkzeug')
    logger.propagate = False

    if get_conf(config, 'Logging.Requests.enabled', False) == False:
        logger.addHandler(NullHandler())
        return

    level_string = get_conf(config, 'Logging.Requests.loglevel', 'info').upper()
    numeric_level = getattr(logging, level_string, None)
    logger.setLevel(numeric_level)

    config_filepath = get_conf(config, 'Logging.Requests.logfile', False)
    if config_filepath == 'default':
        config_filepath = os.path.join(location, 'dagobah_requests.log')
    config_filepath = os.path.expanduser(config_filepath) if config_filepath else None

    if config_filepath:
        file_logger = logging.FileHandler(config_filepath)
        file_logger.setLevel(logging.DEBUG)
        logger.addHandler(file_logger)

    if get_conf(config, 'Logging.Requests.log_to_stdout'):
        stdout_logger = logging.StreamHandler(sys.stdout)
        stdout_logger.setLevel(logging.DEBUG)
        logger.addHandler(stdout_logger)

def configure_app():
    app.debug = get_conf(config, 'Dagobahd.debug', False)
    app.secret_key = get_conf(config, 'Dagobahd.app_secret', 'default_secret')
    app.config['LOGIN_DISABLED'] = get_conf(config,
                                            'Dagobahd.auth_disabled',
                                            False)
    app.config['APP_PASSWORD'] = get_conf(config,
                                          'Dagobahd.password', 'dagobah')

    app.config['AUTH_RATE_LIMIT'] = 30
    app.config['AUTH_ATTEMPTS'] = []
    app.config['APP_HOST'] = get_conf(config, 'Dagobahd.host', '127.0.0.1')
    app.config['APP_PORT'] = get_conf(config, 'Dagobahd.port', '9000')

    configure_requests_logger(config, app)

    login_manager.init_app(app)


def get_conf(config, path, default=None):
    current = config
    for level in path.split('.'):
        if level not in current:
            msg = 'Defaulting missing config key %s to %s' % (path, default)
            print msg
            logging.warning(msg)
            return default
        current = current[level]
    return current


def init_dagobah(testing=False):

    init_core_logger(location, config)

    backend = get_backend(config)
    event_handler = configure_event_hooks(config)
    ssh_config = get_conf(config, 'Dagobahd.ssh_config', '~/.ssh/config')

    if not os.path.isfile(os.path.expanduser(ssh_config)):
        logging.warn("SSH config doesn't exist, no remote hosts will be listed")

    dagobah = Dagobah(backend, event_handler, ssh_config)
    known_ids = [id for id in backend.get_known_dagobah_ids()
                 if id != dagobah.dagobah_id]
    if len(known_ids) > 1:
        # need a way to handle this intelligently through config
        raise ValueError('could not infer dagobah ID, ' +
                         'multiple available in backend')

    if known_ids:
        dagobah.from_backend(known_ids[0])

    return dagobah


def configure_event_hooks(config):
    """ Returns an EventHandler instance with registered hooks. """

    def print_event_info(**kwargs):
        print kwargs.get('event_params', {})

    def job_complete_email(email_handler, **kwargs):
        email_handler.send_job_completed(kwargs['event_params'])

    def job_failed_email(email_handler, **kwargs):
        email_handler.send_job_failed(kwargs['event_params'])

    def task_failed_email(email_handler, **kwargs):
        email_handler.send_task_failed(kwargs['event_params'])

    handler = EventHandler()

    email_handler = get_email_handler(get_conf(config, 'Dagobahd.email', None),
                                      get_conf(config, 'Email', {}))

    if (email_handler and
        get_conf(config, 'Email.send_on_success', False) == True):
        handler.register('job_complete', job_complete_email, email_handler)

    if (email_handler and
        get_conf(config, 'Email.send_on_failure', False) == True):
        handler.register('job_failed', job_failed_email, email_handler)
        handler.register('task_failed', task_failed_email, email_handler)

    return handler


def init_core_logger(location, config):
    """ Initialize the logger with settings from config. """

    logger = logging.getLogger('dagobah')
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    if get_conf(config, 'Logging.Core.enabled', False) == False:
        logger.addHandler(NullHandler())
        return

    config_filepath = get_conf(config, 'Logging.Core.logfile', 'default')
    if config_filepath == 'default':
        config_filepath = os.path.join(location, 'dagobah.log')
    config_filepath = os.path.expanduser(config_filepath) if config_filepath else None

    level_string = get_conf(config, 'Logging.Core.loglevel', 'info').upper()
    numeric_level = getattr(logging, level_string, None)

    basic_config_kwargs = {'level': numeric_level,
                           'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'}
    if config_filepath:
        basic_config_kwargs['filename'] = config_filepath
    else:
        basic_config_kwargs['stream'] = open(os.devnull, 'w')
    logging.basicConfig(**basic_config_kwargs)

    if get_conf(config, 'Logging.Core.log_to_stdout'):
        root = logging.getLogger()
        stdout_logger = logging.StreamHandler(sys.stdout)
        stdout_logger.setLevel(logging.DEBUG)
        stdout_logger.setFormatter(formatter)
        root.addHandler(stdout_logger)

    if config_filepath:
        print 'Logging output to %s' % config_filepath
    logging.info('Core logger initialized at level %s' % level_string)



def get_backend(config):
    """ Returns a backend instance based on the Daemon config file. """

    backend_string = get_conf(config, 'Dagobahd.backend', None)

    if backend_string is None:
        from ..backend.base import BaseBackend
        return BaseBackend()

    elif backend_string.lower() == 'mongo':
        backend_kwargs = {}
        for conf_kwarg in ['host', 'port', 'db',
                           'dagobah_collection', 'job_collection',
                           'log_collection']:
            backend_kwargs[conf_kwarg] = get_conf(config,
                                                  'MongoBackend.%s' % conf_kwarg)
        backend_kwargs['port'] = int(backend_kwargs['port'])

        try:
            from ..backend.mongo import MongoBackend
        except:
            raise ImportError('Could not initialize the MongoDB Backend. Are you sure' +
                              ' the optional drivers are installed? If not, try running ' +
                              '"pip install pymongo" to install them.')
        return MongoBackend(**backend_kwargs)

    raise ValueError('unknown backend type specified in conf')


@app.route('/favicon.ico')
def favicon_redirect():
    return send_from_directory(os.path.join(app.root_path,
                                            'static', 'img'),
                               'favicon.ico',
                               mimetype='image/vnd.microsoft.icon')


config = get_config_file()
dagobah = init_dagobah()
app.config['dagobah'] = dagobah
configure_app()



from .daemon import app, login_manager
from .auth import *
from .api import *
from .views import *

def daemon_entrypoint():
    # TODO: the Flask reloader causes multiple Dagobah instances to get created
    # with two schedulers. Need a fix to reenable Flask reloading.
    print 'Starting app on %s:%s' % (app.config['APP_HOST'],
                                     app.config['APP_PORT'])
    app.run(host=app.config['APP_HOST'], port=app.config['APP_PORT'],
            use_reloader=False)

if __name__ == '__main__':
    daemon_entrypoint()



""" Common email class to base specific templates off of. """

import os
import smtplib
import socket
import email.utils

import jinja2

class EmailTemplate(object):

    def __init__(self, **kwargs):

        self.location = os.path.realpath(os.path.join(os.getcwd(),
                                                      os.path.dirname(__file__)))

        self.formatters = {'{HOSTNAME}': socket.gethostname}
        for kwarg, value in kwargs.iteritems():
            setattr(self, kwarg, value)
        self.from_address = self._apply_formatters(self.from_address)
        self.message = None


    def send_job_completed(self, data):
        raise NotImplementedError()


    def send_task_failed(self, data):
        raise NotImplementedError()


    def send_job_failed(self, data):
        raise NotImplementedError()


    def _construct_and_send(self, subject):
        self._address_message()
        self._set_subject(subject)
        self._send_message()


    def _apply_formatters(self, value):
        new_value = value
        for formatter, call in self.formatters.iteritems():
            new_value = new_value.replace(formatter, call().strip())
        return new_value


    def _address_message(self):
        email_addr = self.from_address if self.user is None else self.user

        self.message['From'] = email.utils.formataddr((self.from_address, email_addr))
        self.message['To'] = ','.join(self.recipients)


    def _set_subject(self, subject):
        self.message['Subject'] = subject


    def _send_message(self):
        s = smtplib.SMTP(self.host, self.port)
        if self.use_tls:
            s.ehlo()
            s.starttls()
            s.ehlo

        if getattr(self, 'auth_required', True):  #Preserve backward compatibility
            s.login(self.user, self.password)

        s.sendmail(self.message['From'],
                   self.recipients,
                   self.message.as_string())


    def _get_template(self, template_name, template_file):
        """ Returns a Jinja2 template of the specified file. """
        template = os.path.join(self.location, 'templates',
                                template_name, template_file)
        return jinja2.Template(open(template).read())



from setuptools import setup, find_packages

setup(name='nagobah',
      version='0.2.2',
      description='Simple DAG-based job scheduler, expanded batch mode by xuanxuan',
      url='https://github.com/littlegump/dagobah',
      download_url="https://github.com/littlegump/dagobah/tarball/0.1.5",
      author='Travis Thieman & a little bit with xuanxuan',
      author_email='travis.thieman@gmail.com; 13060404095@163.com',
      license='WTFPL',
      packages=find_packages(),
      package_data={'dagobah': ['email/templates/basic/*',
                                'daemon/static/css/*',
                                'daemon/static/js/*',
                                'daemon/static/img/*',
                                'daemon/static/lib/*.js',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/*.js',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/*.ks',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/affix/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/alert/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/button/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/carousel/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/collapse/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/dropdown/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/modal/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/popover/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/scrollspy/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/tab/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/tooltip/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/transition/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/typeahead/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/extras/fontawesome/font/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/retina/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/tinygrowl/*',
                                'daemon/templates/*',
                                'daemon/dagobahd.yml']},
      install_requires=['croniter==0.3.3',
                        'python-dateutil==2.1',
                        'pyyaml==3.10',
                        'flask==0.10.1',
                        'premailer==1.13',
                        'flask-login==0.2.6',
                        'semantic_version==2.3.0',
                        'paramiko==1.16.0',
                        'py-dag==1.0.0'],
      test_suite='nose.collector',
      tests_require=['nose', 'pymongo'],
      entry_points={
          'console_scripts': [
              'dagobahd = dagobah.daemon.app:daemon_entrypoint',
              'echo_dagobah_conf = dagobah:print_standard_conf',
              'nagobah = dagobah.nagobah:main',
              ],
          },
      zip_safe=False)
